/*
 * Copyright (C) 2013 Canonical, Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3, as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
 * SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// local
#include "application.h"
#include "application_manager.h"
#include "desktopfilereader.h"

// unity-mir
#include "logging.h"

// mir
#include <mir/shell/application_session.h>

Application::Application(DesktopFileReader* desktopData, qint64 pid,
                         Application::Stage stage, Application::State state)
    : m_desktopData(desktopData)
    , m_pid(pid)
    , m_stage(stage)
    , m_state(state)
    , m_focus(false)
    , m_fullscreen(false)
{
    DASSERT(desktopData != NULL);
    DLOG("Application::Application (this=%p, desktopData=%p, pid=%lld, stage=%d, state=%d",
         this, desktopData, pid, static_cast<int>(stage), static_cast<int>(state));
}

Application::~Application()
{
    DLOG("Application::~Application");
    delete m_desktopData;
}

QString Application::desktopFile() const
{
    return m_desktopData->file();
}

QString Application::appId() const
{
    return m_desktopData->file();
}

QString Application::name() const
{
    return m_desktopData->name();
}

QString Application::comment() const
{
    return m_desktopData->comment();
}

QString Application::icon() const
{
    return m_desktopData->icon();
}

QString Application::exec() const
{
    return m_desktopData->exec();
}

Application::Stage Application::stage() const
{
    return m_stage;
}

Application::State Application::state() const
{
    return m_state;
}

bool Application::focus() const
{
    return m_focus;
}

bool Application::fullscreen() const
{
    return m_fullscreen;
}

std::shared_ptr<mir::shell::ApplicationSession> Application::session() const
{
    return m_session;
}

void Application::setSession(const std::shared_ptr<mir::shell::ApplicationSession>& session)
{
    // TODO(greyback) what if called with new surface?
    m_session = session;
}

void Application::setSessionName(const QString& name)
{
    DLOG("Application::setSessionName (this=%p, name=%s)", this, name.toLatin1().constData());
    if (m_session) {
        LOG("Application::setSessionName should not be called once session exists");
        return;
    }
    m_sessionName = name;
}

void Application::setStage(Application::Stage stage)
{
    DLOG("Application::setStage (this=%p, stage=%d)", this, static_cast<int>(stage));
    if (m_stage != stage) {
        m_stage = stage;
        emit stageChanged();
    }
}

void Application::setState(Application::State state)
{
    DLOG("Application::setState (this=%p, state=%d)", this, static_cast<int>(state));
    if (m_state != state) {
        m_state = state;
        emit stateChanged();
    }
}

void Application::setFocus(bool focus)
{
    DLOG("Application::setFocus (this=%p, focus=%s)", this, focus ? "yes" : "no");
    if (m_focus != focus) {
        m_focus = focus;
        emit focusChanged();
    }
}

void Application::setFullscreen(bool fullscreen)
{
    DLOG("Application::setFullscreen (this=%p, fullscreen=%s)", this, fullscreen ? "yes" : "no");
    if (m_fullscreen != fullscreen) {
        m_fullscreen = fullscreen;
        emit fullscreenChanged();
    }
}
